import copy
from dataclasses import asdict
import logging
from typing import List
import numpy as np
import torch

import torch
from tqdm import tqdm

from peacasso.datamodel import GeneratorConfig
from peacasso.utils import lerp, preprocess_image, preprocess_mask, slerp
from peacasso.pipelines import StableDiffusionPipeline

logger = logging.getLogger("peacasso")


class Runner:
    """Class to image generation applications for Peacasso"""

    def __init__(
        self,
    ) -> None:
        pass

    def get_latents(self, config: GeneratorConfig, pipe: StableDiffusionPipeline) -> torch.Tensor:
        """Get latent vector"""
        mask = None
        latents_dtype = pipe.text_encoder.dtype
        if config.seed is not None and config.seed != "":
            generator = torch.Generator(device=pipe.device).manual_seed(config.seed)
        else:
            # create a random seed and use it to create a generator
            seed = torch.randint(0, 2 ** 32, (1,)).item()
            generator = torch.Generator(device=pipe.device).manual_seed(seed)

        batch_size = 1 if (
            isinstance(
                config.prompt, str) or (
                config.use_prompt_weights and config.prompt_weights is not None)) else len(
            config.prompt)
        # get latent vector
        if config.init_image is None:  # text prompt mode
            if config.height % 8 != 0 or config.width % 8 != 0:
                raise ValueError(
                    f"`height` and `width` have to be divisible by 8 but are {config.height} and {config.width}."
                )
            # get the intial random noise
            latents_shape = (
                batch_size * config.num_images,
                pipe.unet.in_channels,
                config.height // 8,
                config.width // 8)
            if config.latents is None:
                if pipe.device.type == "mps":
                    # randn does not work reproducibly on mps
                    latents = torch.randn(
                        latents_shape,
                        generator=generator,
                        device="cpu",
                        dtype=latents_dtype).to(
                        pipe.device)
                else:
                    latents = torch.randn(
                        latents_shape,
                        generator=generator,
                        device=pipe.device,
                        dtype=latents_dtype)
            else:
                if config.latents.shape != latents_shape:
                    raise ValueError(
                        f"Unexpected latents shape, got {config.latents.shape}, expected {latents_shape}")
                latents = config.latents.to(pipe.device)
        else:  # image prompt mode
            if config.strength < 0 or config.strength > 1:
                raise ValueError(
                    f"The value of strength should in [0.0, 1.0] but is {config.strength}"
                )
            if not isinstance(config.init_image, torch.FloatTensor):
                init_image = preprocess_image(config.init_image)

            init_image = init_image.to(device=pipe.device, dtype=latents_dtype)
            init_latent_dist = pipe.vae.encode(init_image).latent_dist
            init_latents = init_latent_dist.sample(generator=generator)
            init_latents = 0.18215 * init_latents

            # expand init_latents for batch_size
            init_latents = torch.cat([init_latents] * (batch_size * config.num_images), dim=0)

            # handle mask if provided
            if init_image is not None and config.mask_image is not None:
                if not isinstance(config.mask_image, torch.FloatTensor):
                    mask_image = preprocess_mask(config.mask_image)
                mask_image = mask_image.to(device=pipe.device, dtype=latents_dtype)
                mask = torch.cat([mask_image] * batch_size * config.num_images)

                # check sizes
                if not mask.shape == init_latents.shape:
                    raise ValueError(
                        f"The mask {str(mask.shape)} and init_latents {str(init_latents.shape)} should be the same size!")

            latents = init_latents
        return latents, mask

    def get_text_embedding(self, config: GeneratorConfig, pipe: StableDiffusionPipeline) -> None:
        """Get text embedding"""
        # get prompt text embeddings

        if config.text_embeddings is not None:
            return config.text_embeddings

        if isinstance(config.prompt, str):
            batch_size = 1
        elif isinstance(config.prompt, list):
            batch_size = len(config.prompt)
        else:
            raise ValueError(
                f"`prompt` has to be of type `str` or `list` but is {type(config.prompt)}"
            )
        text_inputs = pipe.tokenizer(
            config.prompt,
            padding="max_length",
            max_length=pipe.tokenizer.model_max_length,
            return_tensors="pt",
            truncation=True,
        )
        text_input_ids = text_inputs.input_ids

        # provide warning if prompt is too long for the model
        if text_input_ids.shape[-1] > pipe.tokenizer.model_max_length:
            removed_text = pipe.tokenizer.batch_decode(
                text_input_ids[:, pipe.tokenizer.model_max_length:])
            logger.warning(
                "The following part of your input was truncated because CLIP can only handle sequences up to"
                f" {pipe.tokenizer.model_max_length} tokens: {removed_text}")
            text_input_ids = text_input_ids[:, : pipe.tokenizer.model_max_length]
        text_embeddings = pipe.text_encoder(text_input_ids.to(pipe.device))[0]

        # apply prompt weights if provided
        if config.use_prompt_weights and config.prompt_weights is not None:
            logging.info("applying prompt_weights" + str(config.prompt_weights))
            if len(config.prompt_weights) != len(config.prompt):
                raise ValueError("Length of prompt should be same as weights.")
            else:
                for i in range(len(config.prompt_weights)):
                    text_embeddings[i] = text_embeddings[i] * config.prompt_weights[i]
                # sum embeddings
                text_embeddings = torch.sum(text_embeddings, dim=0, keepdim=True)
            text_embeddings = torch.mean(text_embeddings, dim=0).unsqueeze(0)
            batch_size = text_embeddings.shape[0]

        # duplicate text embeddings for each generation per prompt, using mps friendly method
        bs_embed, seq_len, _ = text_embeddings.shape
        text_embeddings = text_embeddings.repeat(1, config.num_images, 1)
        text_embeddings = text_embeddings.view(bs_embed * config.num_images, seq_len, -1)

        do_classifier_free_guidance = config.guidance_scale > 1.0
        if do_classifier_free_guidance:
            uncond_tokens: List[str]
            if config.negative_prompt is None:
                uncond_tokens = [""]
            elif not isinstance(config.prompt, type(config.negative_prompt)):
                raise TypeError(
                    f"`negative_prompt` should be the same type to `prompt`, but got {type(config.negative_prompt)} !="
                    f" {type(config.prompt)}.")
            elif isinstance(config.negative_prompt, str):
                uncond_tokens = [config.negative_prompt]
            elif batch_size != len(config.negative_prompt):
                raise ValueError(
                    f"`negative_prompt`: {config.negative_prompt} has batch size {len(config.negative_prompt)}, but `prompt`:"
                    f" {config.prompt} has batch size {batch_size}. Please make sure that passed `negative_prompt` matches"
                    " the batch size of `prompt`.")
            else:
                uncond_tokens = config.negative_prompt

            max_length = text_input_ids.shape[-1]
            uncond_input = pipe.tokenizer(
                uncond_tokens,
                padding="max_length",
                max_length=max_length,
                truncation=True,
                return_tensors="pt",
            )
            uncond_embeddings = pipe.text_encoder(uncond_input.input_ids.to(pipe.device))[0]

            # duplicate unconditional embeddings for each generation per prompt, using
            # mps friendly method
            seq_len = uncond_embeddings.shape[1]
            uncond_embeddings = uncond_embeddings.repeat(batch_size, config.num_images, 1)
            uncond_embeddings = uncond_embeddings.view(
                batch_size * config.num_images, seq_len, -1)

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings into a single batch
            # to avoid doing two forward passes
            text_embeddings = torch.cat([uncond_embeddings, text_embeddings])

        return text_embeddings

    def run(self, config: GeneratorConfig, pipe: StableDiffusionPipeline) -> None:
        """Run application"""
        if config.application is None or "type" not in config.application:
            text_embeddings = self.get_text_embedding(config, pipe)
            latents, mask = self.get_latents(config, pipe)

            config = asdict(config)
            config["text_embeddings"] = text_embeddings
            config["latents"] = latents
            config["mask"] = mask
            return pipe(**config)

        else:
            if "type" in config.application:
                app_type = config.application["type"]
                app_config = config.application["config"]
                # slerp = app_config["function"] if "function" in app_config else lerp
                if app_type == "interpolate":
                    config.num_images = 1  # batch size must be 1 for interpolation
                    if app_config is None:
                        raise ValueError(
                            "application_config must be provided for interpolate application")
                    else:
                        if "num_steps" not in app_config:
                            raise ValueError(
                                "num_steps must be provided for interpolate configuration")
                        else:
                            num_steps = app_config["num_steps"]
                            start_config, end_config = copy.copy(config), copy.copy(config)
                            # get start and end latents. note that these will be based on seed if
                            # provided, images if provided and masks if provided
                            if "seed" in app_config:
                                start_config.seed = app_config["seed"]["start"]
                                end_config.seed = app_config["seed"]["end"]
                            if "image" in app_config:
                                start_config.init_image = app_config["image"]["start"]
                                end_config.init_image = app_config["image"]["end"]
                            start_latents, start_mask = self.get_latents(start_config, pipe)
                            end_latents, end_mask = self.get_latents(end_config, pipe)
                            if "prompt" in app_config:
                                start_config.prompt = app_config["prompt"]["start"]
                                end_config.prompt = app_config["prompt"]["end"]
                                start_text_embeddings = self.get_text_embedding(start_config, pipe)
                                end_text_embeddings = self.get_text_embedding(end_config, pipe)

                            # interpolate between start and end configs
                            interpolation_results = []
                            text_embeddings = self.get_text_embedding(config, pipe)
                            conf_dict = asdict(config)
                            for t in tqdm(np.linspace(0, 1, num_steps)):
                                latents = slerp(float(t), start_latents, end_latents)
                                if "prompt" in app_config:
                                    text_embeddings = slerp(
                                        float(t), start_text_embeddings, end_text_embeddings)
                                # mask = slerp(float(t), start_mask, end_mask)
                                conf_dict["text_embeddings"] = text_embeddings
                                conf_dict["latents"] = latents
                                conf_dict["mask"] = end_mask
                                result = pipe(**conf_dict)
                                interpolation_results.append(result)
                            return interpolation_results
                else:
                    raise ValueError(f"application type {app_type} not supported")



# based on
# https://github.com/huggingface/diffusers/tree/main/src/diffusers/pipelines/stable_diffusion
import inspect
import logging
import time
from typing import Callable, List, Optional, Union
import PIL
import torch
import numpy as np

from tqdm.auto import tqdm
from transformers import CLIPFeatureExtractor, CLIPTextModel, CLIPTokenizer

# from ...models import AutoencoderKL, UNet2DConditionModel
# from ...pipeline_utils import DiffusionPipeline
# from ...schedulers import DDIMScheduler, LMSDiscreteScheduler, PNDMScheduler
# from .safety_checker import StableDiffusionSafetyChecker

from diffusers.models import AutoencoderKL, UNet2DConditionModel
from diffusers.pipeline_utils import DiffusionPipeline
from diffusers.schedulers import DDIMScheduler, LMSDiscreteScheduler, PNDMScheduler
from diffusers.pipelines.stable_diffusion.safety_checker import (
    StableDiffusionSafetyChecker,
)

logger = logging.getLogger(__name__)


def preprocess(image):
    w, h = image.size
    w, h = map(lambda x: x - x % 64, (w, h))  # resize to integer multiple of 32
    image = image.resize((w, h), resample=PIL.Image.LANCZOS)
    image = np.array(image).astype(np.float32) / 255.0
    image = image[None].transpose(0, 3, 1, 2)
    image = torch.from_numpy(image)
    return 2.0 * image - 1.0


def preprocess_mask(mask):
    w, h = mask.size
    w, h = map(lambda x: x - x % 32, (w, h))  # resize to integer multiple of 32
    mask = mask.resize((w // 8, h // 8), resample=PIL.Image.NEAREST)
    mask = np.array(mask).astype(np.float32) / 255.0
    mask = np.tile(mask, (4, 1, 1))
    mask = mask[None].transpose(0, 1, 2, 3)  # what does this step do?
    # mask = 1 - mask  # repaint white, keep black
    mask = torch.from_numpy(mask)
    return mask


def decode_image(
    latents,
    vae,
):
    latents = 1 / 0.18215 * latents
    image = vae.decode(latents.to(vae.dtype)).sample
    image = (image / 2 + 0.5).clamp(0, 1)
    image = image.cpu().permute(0, 2, 3, 1).numpy()
    return image


class StableDiffusionPipeline(DiffusionPipeline):
    def __init__(
        self,
        vae: AutoencoderKL,
        text_encoder: CLIPTextModel,
        tokenizer: CLIPTokenizer,
        unet: UNet2DConditionModel,
        scheduler: Union[DDIMScheduler, PNDMScheduler, LMSDiscreteScheduler],
        feature_extractor: CLIPFeatureExtractor,
        safety_checker: StableDiffusionSafetyChecker,
    ):
        super().__init__()
        self.register_modules(
            vae=vae,
            text_encoder=text_encoder,
            tokenizer=tokenizer,
            unet=unet,
            scheduler=scheduler,
            feature_extractor=feature_extractor,
            safety_checker=safety_checker,
        )

    @torch.no_grad()
    def __call__(
        self,
        prompt: Union[str, List[str]],
        height: Optional[int] = 512,
        width: Optional[int] = 512,
        num_inference_steps: Optional[int] = 50,
        guidance_scale: Optional[float] = 7.5,
        eta: Optional[float] = 0.0,
        generator: Optional[torch.Generator] = None,
        strength: float = 0.8,
        init_image: Union[torch.FloatTensor, PIL.Image.Image] = None,
        seed: Optional[int] = 2147483647,
        attention_slice: Optional[Union[str, int]] = "auto",
        mask_image: Union[torch.FloatTensor, PIL.Image.Image] = None,
        num_images: Optional[int] = 1,
        negative_prompt: Optional[Union[str, List[str]]] = None,
        latents: Optional[torch.FloatTensor] = None,
        callback: Optional[Callable[[int, int, torch.FloatTensor, PIL.Image.Image], None]] = None,
        callback_steps: Optional[int] = 1,
        return_intermediates: Optional[bool] = False,
        prompt_weights: Optional[Union[List[float], float]] = None,
        use_prompt_weights: Optional[bool] = False,
        text_embeddings: Optional[torch.FloatTensor] = None,
        text_input_ids: Optional[torch.LongTensor] = None,
        mask: Optional[torch.FloatTensor] = None,
        filter_nsfw: Optional[bool] = False,
        **kwargs,
    ):

        intermediate_images = []
        if (callback_steps is None) or (callback_steps is not None and (
                not isinstance(callback_steps, int) or callback_steps <= 0)):
            raise ValueError(
                f"`callback_steps` has to be a positive integer but is {callback_steps} of type"
                f" {type(callback_steps)}."
            )

        # set seed
        if seed is not None and seed != "":
            generator = torch.Generator(device=self.device).manual_seed(seed)
        else:
            # create a random seed and use it to create a generator
            seed = torch.randint(0, 2 ** 32, (1,)).item()
            generator = torch.Generator(device=self.device).manual_seed(seed)
        # enable attention slicing if attention_slice is not None
        if attention_slice:
            if attention_slice == "auto":
                self.unet.set_attention_slice(self.unet.config.attention_head_dim // 2)
            else:
                self.unet.set_attention_slice(attention_slice)

        start_time = time.time()
        batch_size = 1 if (
            isinstance(
                prompt, str) or (
                use_prompt_weights and prompt_weights is not None)) else len(prompt)

        self.scheduler.set_timesteps(num_inference_steps)
        timesteps_tensor = self.scheduler.timesteps.to(self.device)

        # here `guidance_scale` is defined analog to the guidance weight `w` of equation (2)
        # of the Imagen paper: https://arxiv.org/pdf/2205.11487.pdf . `guidance_scale = 1`
        # corresponds to doing no classifier free guidance.
        do_classifier_free_guidance = guidance_scale > 1.0

        if init_image:
            init_latents_orig = latents
            offset = self.scheduler.config.get("steps_offset", 0)
            init_timestep = int(num_inference_steps * strength) + offset
            init_timestep = min(init_timestep, num_inference_steps)

            timesteps = self.scheduler.timesteps[-init_timestep]
            timesteps = torch.tensor([timesteps] * batch_size * num_images, device=self.device)

            # add noise to latents using the timesteps
            noise = torch.randn(
                latents.shape,
                generator=generator,
                device=self.device,
                dtype=latents.dtype)
            latents = self.scheduler.add_noise(latents, noise, timesteps)
            # t_start = max(num_inference_steps - init_timestep + offset, 0)
            # timesteps = self.scheduler.timesteps[t_start:].to(self.device)

        # prepare extra kwargs for the scheduler step, since not all schedulers have the same signature
        # eta (η) is only used with the DDIMScheduler, it will be ignored for other schedulers.
        # eta corresponds to η in DDIM paper: https://arxiv.org/abs/2010.02502
        # and should be between [0, 1]
        accepts_eta = "eta" in set(inspect.signature(self.scheduler.step).parameters.keys())
        extra_step_kwargs = {}
        if accepts_eta:
            extra_step_kwargs["eta"] = eta

        for i, t in enumerate(self.progress_bar(timesteps_tensor)):
            # expand the latents if we are doing classifier free guidance
            latent_model_input = torch.cat(
                [latents] * 2) if do_classifier_free_guidance else latents

            latent_model_input = self.scheduler.scale_model_input(latent_model_input, t)
            # predict the noise residual
            noise_pred = self.unet(
                latent_model_input,
                t,
                encoder_hidden_states=text_embeddings).sample

            # perform guidance
            if do_classifier_free_guidance:
                noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                noise_pred = noise_pred_uncond + guidance_scale * \
                    (noise_pred_text - noise_pred_uncond)

            # compute the previous noisy sample x_t -> x_t-1
            latents = self.scheduler.step(noise_pred, t, latents, **extra_step_kwargs).prev_sample
            if init_image is not None and mask is not None:
                init_latents_proper = self.scheduler.add_noise(init_latents_orig, noise, t)
                latents = (init_latents_proper * mask) + (latents * (1 - mask))

            decoded_image = None
            if return_intermediates:
                decoded_image = self.numpy_to_pil(decode_image(latents, self.vae))
                intermediate_images.append(decoded_image)

            # call the callback, if provided
            if callback is not None and i % callback_steps == 0:
                callback(i, t, latents, decoded_image)

        # scale and decode the image latents with vae
        image = decode_image(latents, self.vae)
        has_nsfw_concept = None
        if filter_nsfw:

            safety_cheker_input = self.feature_extractor(
                self.numpy_to_pil(image), return_tensors="pt"
            ).to(self.device)
            image, has_nsfw_concept = self.safety_checker(
                images=np.asarray(image), clip_input=safety_cheker_input.pixel_values.to(
                    text_embeddings.dtype))
        image = self.numpy_to_pil(image)

        return {
            "images": image,
            "intermediates": intermediate_images,
            "nsfw_content_detected": has_nsfw_concept,
            "time": time.time() - start_time,
            "seed": seed,
        }



import copy
from dataclasses import asdict
import logging
from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
import os
from peacasso.generator import ImageGenerator
from fastapi.middleware.cors import CORSMiddleware
from peacasso.datamodel import ModelConfig, WebRequestData
from peacasso.utils import base64_to_pil, pil_to_base64, sanitize_config
import os
import traceback

logger = logging.getLogger("peacasso")

assert os.environ.get("HF_API_TOKEN") is not None, "HF_API_TOKEN not set"

# load model using env variables
model_config = ModelConfig(
    model=os.environ.get("PEACASSO_MODEL", "runwayml/stable-diffusion-v1-5"),
    revision=os.environ.get("PEACASSO_REVISION", "fp16"),
    device=os.environ.get("PEACASSO_DEVICE", "cuda:0"),
    token=os.environ.get("HF_API_TOKEN")
)
logger.info(
    ">> Loading Imagenerator pipeline with config. " + str(model_config))
generator =  ImageGenerator(model_config)
logger.info(">> Imagenerator pipeline loaded.")

app = FastAPI()
# allow cross origin requests for testing on localhost: 800 * ports only
app.add_middleware(
    CORSMiddleware,
    allow_origins=["http://localhost:8000", "http://localhost:8001"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
api = FastAPI(root_path="/api")
app.mount("/api", api)


root_file_path = os.path.dirname(os.path.abspath(__file__))
static_folder_root = os.path.join(root_file_path, "ui")
files_static_root = os.path.join(root_file_path, "files/")

os.makedirs(files_static_root, exist_ok=True)

# mount peacasso front end UI files
app.mount("/", StaticFiles(directory=static_folder_root, html=True), name="ui")
api.mount("/files", StaticFiles(directory=files_static_root, html=True), name="files")


@api.post("/generate")
def generate(request: WebRequestData) -> str:
    """Generate an image given some prompt"""

    result = None
    try:
        if request.type == "generate":
            prompt_config = request.config
            sanitized_config = asdict(sanitize_config(copy.deepcopy(prompt_config))) 
            if prompt_config.init_image:
                prompt_config.init_image, _ = base64_to_pil(prompt_config.init_image)
            if prompt_config.mask_image:
                _, prompt_config.mask_image = base64_to_pil(prompt_config.mask_image)
            response = None 
            try: 
                result = generator.generate(prompt_config)
                images = []
                for image in result["images"]: # convert pil image to base64 and prepend with data uri
                    images.append("data:image/png;base64, " + pil_to_base64(image))
                result["images"] = images  
                response = {"status": True, "status_message": "success", "config": sanitized_config, "result": result} 
                
            except Exception as e: 
                traceback.print_exc()
                response =  {"status": False, "status_message": str(e), "config": sanitized_config, "result": result}
            return response
        else:
            return {"status": False, "status_message": "invalid request type"}            
    except Exception as e:
        traceback.print_exc()
        return {"status": False, "status_message": str(e)}
     


from dataclasses import asdict
import logging
from PIL import Image
from typing import List, Optional
# from diffusers import StableDiffusionPipeline

import os
import torch
import time

from peacasso.datamodel import GeneratorConfig, ModelConfig
from peacasso.applications import Runner
from peacasso.pipelines import StableDiffusionPipeline
from diffusers import DPMSolverMultistepScheduler

logger = logging.getLogger(__name__)


class ImageGenerator:
    """Class to handle generating image from prompt"""

    def __init__(
        self,
        model_config: ModelConfig
    ) -> None:

        self.create_model(model_config)

    def create_model(self, model_config: ModelConfig): 
        if model_config.device == None:
            if torch.cuda.is_available():
                device = "cuda"
            elif torch.backends.mps.is_available():
                device = "mps"
            else:
                device = "cpu"
        else:
            device = model_config.device 
            
        self.pipe = StableDiffusionPipeline.from_pretrained(
            model_config.model,
            revision=model_config.revision,
            torch_dtype=torch.float16 if "cuda" in device else torch.float32,
            use_auth_token=model_config.token
        ).to(device)
        self.pipe.scheduler = DPMSolverMultistepScheduler.from_config(self.pipe.scheduler.config)

        self.runner = Runner()

    def generate(self, config: GeneratorConfig) -> Image:
        """Generate image from prompt"""
        return self.runner.run(config, self.pipe)

    def reload(self, model_config) -> None:
        """Clear pipe models from memory, reload pipe with new parameters"""
        del self.pipe
        torch.cuda.empty_cache()
        self.create_model(model_config)

    def list_cuda(self) -> List[int]:
        """List available cuda devices
        Returns:
            List[int]: List of available cuda devices
        """
        available_gpus = [i for i in range(torch.cuda.device_count())]
        return available_gpus
